version='0.3.1.dev4'
git_hash='1377e7f'
git_branch='jeffra/jit_extensions_v2'
installed_ops={'cpu_adam': False, 'fused_adam': False, 'fused_lamb': False, 'sparse_attn': False, 'transformer': False, 'stochastic_transformer': False, 'utils': False}
compatible_ops={'cpu_adam': True, 'fused_adam': True, 'fused_lamb': True, 'sparse_attn': False, 'transformer': True, 'stochastic_transformer': True, 'utils': True}
torch_info={'version': '1.5', 'cuda_version': '10.2'}
